/*
 *  partbl_sgi.c -- SGI partition table parser
 *
 *  Code extracted from
 *  linux/drivers/block/genhd.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <linux/module.h>

#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/init.h>


static int sgi_partition(struct gendisk *hd, kdev_t dev, unsigned long first_sector)
{
	int i, csum, magic;
	unsigned int *ui, start, blocks, cs;
	struct buffer_head *bh;
	struct sgi_disklabel {
		int magic_mushroom;         /* Big fat spliff... */
		short root_part_num;        /* Root partition number */
		short swap_part_num;        /* Swap partition number */
		char boot_file[16];         /* Name of boot file for ARCS */
		unsigned char _unused0[48]; /* Device parameter useless crapola.. */
		struct sgi_volume {
			char name[8];       /* Name of volume */
			int  block_num;     /* Logical block number */
			int  num_bytes;     /* How big, in bytes */
		} volume[15];
		struct sgi_partition {
			int num_blocks;     /* Size in logical blocks */
			int first_block;    /* First logical block */
			int type;           /* Type of this partition */
		} partitions[16];
		int csum;                   /* Disk label checksum */
		int _unused1;               /* Padding */
	} *label;
	struct sgi_partition *p;
#define SGI_LABEL_MAGIC 0x0be5a941

	MOD_INC_USE_COUNT;
	if(!(bh = bread(dev, 0, get_ptable_blocksize(dev)))) {
		printk("Dev %s: unable to read partition table\n", kdevname(dev));
		MOD_DEC_USE_COUNT;
		return -1;
	}
	label = (struct sgi_disklabel *) bh->b_data;
	p = &label->partitions[0];
	magic = label->magic_mushroom;
	if(be32_to_cpu(magic) != SGI_LABEL_MAGIC) {
		printk("Dev %s SGI disklabel: bad magic %08x\n",
		       kdevname(dev), magic);
		brelse(bh);
		MOD_DEC_USE_COUNT;
		return 0;
	}
	ui = ((unsigned int *) (label + 1)) - 1;
	for(csum = 0; ui >= ((unsigned int *) label);) {
		cs = *ui--;
		csum += be32_to_cpu(cs);
	}
	if(csum) {
		printk("Dev %s SGI disklabel: csum bad, label corrupted\n",
		       kdevname(dev));
		brelse(bh);
		MOD_DEC_USE_COUNT;
		return 0;
	}
	printk(" SGI");
	/* All SGI disk labels have 16 partitions, disks under Linux only
	 * have 15 minor's.  Luckily there are always a few zero length
	 * partitions which we don't care about so we never overflow the
	 * current_minor.
	 */
	for(i = 0; i < 16; i++, p++) {
		blocks = be32_to_cpu(p->num_blocks);
		start  = be32_to_cpu(p->first_block);
		if(!blocks)
			continue;
		partbl_add_partition(hd, current_minor, start, blocks);
		current_minor++;
	}
	printk("\n");
	brelse(bh);
	MOD_DEC_USE_COUNT;
	return 1;
}

static struct partition_parser sgi_partition_parser =
    { NULL, "sgi", sgi_partition };

#ifndef MODULE
__initfunc(void sgi_partbl_init( void ))
{
	(void)register_partbl( &sgi_partition_parser );
}
#else
int init_module(void)
{
	return register_partbl( &sgi_partition_parser );
}

void cleanup_module(void)
{
	(void)unregister_partbl( &sgi_partition_parser );
}
#endif
